import glob
import io
import os

import pygame
from PIL import Image

from nubrain.experiment.global_config import GlobalConfig

global_config = GlobalConfig()
max_img_storage_dimension = global_config.max_img_storage_dimension


def get_all_image_paths(*, image_directory: str):
    """
    Get the path of all images in target directory.
    """
    extensions = ("*.png", "*.jpg", "*.jpeg")
    image_file_paths = []
    for ext in extensions:
        image_file_paths.extend(glob.glob(os.path.join(image_directory, ext)))

    if not image_file_paths:
        print(f"No images found in directory: {image_directory}")
        return []

    print(f"Found {len(image_file_paths)} images.")

    return image_file_paths


def load_and_scale_image(
    *,
    image_file_path: str,
    screen_width: int,
    screen_height: int,
):
    """
    Loads image as pygame object and scale them to fit the screen, if the image is
    larger than the screen (to be used for stimulus presentation).
    """

    # (1) Load image.
    try:
        image = pygame.image.load(image_file_path)
        img_rect = image.get_rect()

        if (screen_width < img_rect.width) or (screen_height < img_rect.height):
            # The image is larger than the screen. Scale it to fit the screen while
            # maintaining the aspect ratio.
            scale_w = screen_width / img_rect.width
            scale_h = screen_height / img_rect.height
            scale = min(scale_w, scale_h)

            new_width = int(img_rect.width * scale)
            new_height = int(img_rect.height * scale)
            image = pygame.transform.smoothscale(image, (new_width, new_height))

        else:
            # The image is not too large for the screen.
            pass

    except pygame.error as e:
        print(f"Error loading or scaling image {image_file_path}: {e}")
        return None

    # (2) Load image category from text file (if it exists).
    try:
        file_path_withou_extension = os.path.splitext(image_file_path)[0]
        path_txt = file_path_withou_extension + ".txt"

        if os.path.isfile(path_txt):
            with open(path_txt, "r", encoding="utf-8") as file:
                image_category = file.read()
            image_category = image_category.strip()
        else:
            image_category = "null"  # Needs to be string for saving to hdf5
    except Exception as e:
        print(f"Error loading image metadata {path_txt}: {e}")
        return None

    image_and_metadata = {
        "image_file_path": image_file_path,
        "image": image,
        "image_category": image_category,
    }

    return image_and_metadata


def load_image_as_bytes(*, image_path: str):
    """
    Load an image file from disk and return it as a bytes object.
    """
    with open(image_path, "rb") as f:
        image_bytes = f.read()
    return image_bytes


def resize_image(*, image_bytes: bytes, return_image_file_extension: bool = False):
    """
    Resize image to maximal size (not used for stimulus presentation, but for logging).
    """
    image = Image.open(io.BytesIO(image_bytes))

    if image.format == "JPEG":
        image_format = "JPEG"
        image_file_extension = ".jpg"
    elif image.format == "PNG":
        image_format = "PNG"
        image_file_extension = ".png"
    elif image.format == "WEBP":
        image_format = "WEBP"
        image_file_extension = ".webp"
    else:
        print(f"Unexpected image format, will use png: {image.format}")
        image_format = "PNG"
        image_file_extension = ".png"

    width, height = image.size

    # Check if resizing is needed.
    if (width > max_img_storage_dimension) or (height > max_img_storage_dimension):
        # Calculate the new size maintaining the aspect ratio.
        if width > height:
            new_width = max_img_storage_dimension
            new_height = int(max_img_storage_dimension * height / width)
        else:
            new_height = max_img_storage_dimension
            new_width = int(max_img_storage_dimension * width / height)

        # Resize the image.
        image = image.resize((new_width, new_height), Image.Resampling.BILINEAR)

    # Convert the image to bytes again.
    image_bytes_resized = io.BytesIO()

    image.save(image_bytes_resized, format=image_format)
    image_bytes_resized = bytearray(image_bytes_resized.getvalue())

    if return_image_file_extension:
        return image_bytes_resized, image_file_extension
    else:
        return image_bytes_resized



import multiprocessing as mp
import os
import random
from time import sleep, time

import pygame
from brainflow.board_shim import BoardIds, BoardShim, BrainFlowInputParams

from nubrain.experiment.data import eeg_data_logging
from nubrain.experiment.global_config import GlobalConfig
from nubrain.image.tools import get_all_image_paths, load_and_scale_image
from nubrain.misc.datetime import get_formatted_current_datetime


def experiment(config: dict):
    # ----------------------------------------------------------------------------------
    # *** Get config

    demo_mode = config["demo_mode"]

    output_directory = config["output_directory"]
    image_directory = config["image_directory"]

    eeg_channel_mapping = config["eeg_channel_mapping"]

    # bandstop_low_cutoff_freq = config["bandstop_low_cutoff_freq"]
    # bandstop_high_cutoff_freq = config["bandstop_high_cutoff_freq"]
    # bandstop_filter_order = config["bandstop_filter_order"]

    # bandpass_low_cutoff_freq = config["bandpass_low_cutoff_freq"]
    # bandpass_high_cutoff_freq = config["bandpass_high_cutoff_freq"]
    # bandpass_filter_order = config["bandpass_filter_order"]

    initial_rest_duration = config["initial_rest_duration"]
    image_duration = config["image_duration"]
    isi_duration = config["isi_duration"]
    inter_block_grey_duration = config["inter_block_grey_duration"]

    n_blocks = config["n_blocks"]
    images_per_block = config["images_per_block"]

    eeg_device_address = config["eeg_device_address"]

    global_config = GlobalConfig()

    # ----------------------------------------------------------------------------------
    # *** Test if output path exists

    if not os.path.isdir(output_directory):
        raise AssertionError(f"Target directory does not exist: {output_directory}")

    current_datetime = get_formatted_current_datetime()
    path_out_data = os.path.join(output_directory, f"eeg_session_{current_datetime}.h5")

    if os.path.isfile(path_out_data):
        raise AssertionError(f"Target file aready exists: {path_out_data}")

    # ----------------------------------------------------------------------------------
    # *** Get image paths

    image_file_paths = get_all_image_paths(image_directory=image_directory)

    if not image_file_paths:
        raise AssertionError(f"Found no images at {image_directory}")
    print(f"Found {len(image_file_paths)} images")

    # ----------------------------------------------------------------------------------
    # *** Prepare EEG measurement

    BoardShim.enable_dev_board_logger()

    if demo_mode:
        board_id = BoardIds.SYNTHETIC_BOARD.value
    else:
        board_id = BoardIds.CYTON_BOARD.value

    params = BrainFlowInputParams()
    params.serial_port = eeg_device_address
    board = BoardShim(board_id, params)

    eeg_board_description = BoardShim.get_board_descr(board_id)
    eeg_sampling_rate = int(eeg_board_description["sampling_rate"])
    eeg_channels = eeg_board_description["eeg_channels"]  # Get EEG channel indices
    marker_channel = eeg_board_description["marker_channel"]

    board.prepare_session()

    print(f"Board: {eeg_board_description['name']}")
    print(f"Sampling Rate: {eeg_sampling_rate} Hz")
    print(f"EEG Channels: {eeg_channels}")

    board.start_stream()

    sleep(0.1)
    board_data = board.get_board_data()

    # Total number of channels, including EEG, marker, and other channels.
    n_channels_total = board_data.shape[0]

    # ----------------------------------------------------------------------------------
    # *** Start data logging subprocess

    data_logging_queue = mp.Queue()

    subprocess_params = {
        "demo_mode": demo_mode,
        "image_directory": image_directory,
        # EEG parameters
        "eeg_board_description": eeg_board_description,
        "eeg_sampling_rate": eeg_sampling_rate,
        "n_channels_total": n_channels_total,
        "eeg_channels": eeg_channels,
        "marker_channel": marker_channel,
        "eeg_channel_mapping": eeg_channel_mapping,
        "eeg_device_address": eeg_device_address,
        # Timing parameters
        "initial_rest_duration": initial_rest_duration,
        "image_duration": image_duration,
        "isi_duration": isi_duration,
        "inter_block_grey_duration": inter_block_grey_duration,
        # Experiment structure
        "n_blocks": n_blocks,
        "images_per_block": images_per_block,
        # Filter parameters
        # "bandstop_low_cutoff_freq": bandstop_low_cutoff_freq,
        # "bandstop_high_cutoff_freq": bandstop_high_cutoff_freq,
        # "bandstop_filter_order": bandstop_filter_order,
        # "bandpass_low_cutoff_freq": bandpass_low_cutoff_freq,
        # "bandpass_high_cutoff_freq": bandpass_high_cutoff_freq,
        # "bandpass_filter_order": bandpass_filter_order,
        # Misc
        # "nubrain_endpoint": nubrain_endpoint,
        # "nubrain_api_key": nubrain_api_key,
        "path_out_data": path_out_data,
        "data_logging_queue": data_logging_queue,
    }

    logging_process = mp.Process(
        target=eeg_data_logging,
        args=(subprocess_params,),
    )

    logging_process.Daemon = True
    logging_process.start()

    # ----------------------------------------------------------------------------------
    # *** Start experiment

    running = True
    while running:
        pygame.init()

        # Get screen dimensions and set up full screen
        screen_info = pygame.display.Info()
        screen_width = screen_info.current_w
        screen_height = screen_info.current_h
        screen = pygame.display.set_mode(
            (screen_width, screen_height), pygame.FULLSCREEN
        )
        pygame.display.set_caption("Image Presentation Experiment")
        pygame.mouse.set_visible(False)  # Hide the mouse cursor

        font = pygame.font.Font(None, 48)  # Basic font for messages

        # Load first image.
        image_and_metadata = None
        while image_and_metadata is None:
            # Select a random image from the full list.
            random_image_file_path = random.choice(image_file_paths)
            image_and_metadata = load_and_scale_image(
                image_file_path=random_image_file_path,
                screen_width=screen_width,
                screen_height=screen_height,
            )

        try:
            # Initial grey screen.
            pygame.time.wait(100)
            screen.fill(global_config.rest_condition_color)
            pygame.display.flip()
            pygame.time.wait(100)
            screen.fill(global_config.rest_condition_color)
            pygame.display.flip()

            # Clear board buffer.
            _ = board.get_board_data()

            # Pause for specified number of milliseconds.
            pygame.time.delay(int(round(initial_rest_duration * 1000.0)))

            # Block loop.
            for idx_block in range(n_blocks):
                print(f"Starting Block {idx_block + 1} out of {n_blocks}")

                # Image loop (within a block).
                for image_count in range(images_per_block):
                    if not running:
                        break  # Check for quit event

                    image_file_path = image_and_metadata["image_file_path"]
                    current_image = image_and_metadata["image"]
                    image_category = image_and_metadata["image_category"]

                    img_rect = current_image.get_rect(
                        center=(screen_width // 2, screen_height // 2)
                    )

                    # Display image. Clear previous screen content.
                    screen.fill(global_config.rest_condition_color)
                    screen.blit(current_image, img_rect)
                    pygame.display.flip()

                    # Start of stimulus presentation.
                    t1 = time()
                    # Insert stimulus start maker into EEG data.
                    board.insert_marker(global_config.stim_start_marker)

                    # Send pre-stimulus board data (to avoid buffer overflow).
                    data_to_queue = {
                        "board_data": board.get_board_data(),
                        "stimulus_data": None,
                    }
                    data_logging_queue.put(data_to_queue)

                    # Time until when to show stimulus.
                    t2 = t1 + image_duration
                    while time() < t2:
                        pass

                    # End of stimulus presentation. Display ISI grey screen.
                    screen.fill(global_config.rest_condition_color)
                    pygame.display.flip()
                    board.insert_marker(global_config.stim_end_marker)
                    t3 = time()

                    # Send data corresponding to stimulus period.
                    stimulus_data = {
                        "stimulus_start_time": t1,
                        "stimulus_end_time": t3,
                        "stimulus_duration_s": t3 - t1,
                        "image_file_path": image_file_path,
                        "image_category": image_category,
                    }
                    data_to_queue = {
                        "board_data": board.get_board_data(),
                        "stimulus_data": stimulus_data,
                    }
                    data_logging_queue.put(data_to_queue)

                    # Load next image.
                    image_and_metadata = None
                    while image_and_metadata is None:
                        # Select a random image from the full list.
                        random_image_file_path = random.choice(image_file_paths)
                        image_and_metadata = load_and_scale_image(
                            image_file_path=random_image_file_path,
                            screen_width=screen_width,
                            screen_height=screen_height,
                        )

                    # Time until when to show grey screen.
                    t4 = t3 + isi_duration
                    while time() < t4:
                        pass

                    # Event handling (allow quitting with ESC or window close).
                    for event in pygame.event.get():
                        if event.type == pygame.QUIT:
                            running = False
                        if event.type == pygame.KEYDOWN:
                            if event.key == pygame.K_ESCAPE:
                                running = False

                    if not running:
                        break

                if not running:
                    break

                # Send post-stimulus board data (to avoid buffer overflow).
                data_to_queue = {
                    "board_data": board.get_board_data(),
                    "stimulus_data": None,
                }
                data_logging_queue.put(data_to_queue)

                # Inter-block grey screen.
                print(f"End of Block {idx_block + 1}. Starting inter-block interval.")
                screen.fill(global_config.rest_condition_color)
                pygame.display.flip()
                # We already waited for the ISI duration, therefore subtract it from the
                # inter block duration. Avoid negative value is case ISI duration is
                # longer than inter block duration.
                remaining_wait = max((inter_block_grey_duration - isi_duration), 0.0)
                pygame.time.delay(int(round(remaining_wait * 1000.0)))

            # End of experiment.
            if running:  # Only show if not quit early
                screen.fill(global_config.rest_condition_color)
                end_text = font.render("Experiment complete.", True, (0.0, 0.0, 0.0))
                text_rect = end_text.get_rect(
                    center=(screen_width // 2, screen_height // 2)
                )
                screen.blit(end_text, text_rect)
                pygame.display.flip()
                pygame.time.wait(500)

            running = False

            # Send final board data.
            data_to_queue = {
                "board_data": board.get_board_data(),
                "stimulus_data": None,
            }
            data_logging_queue.put(data_to_queue)

        except Exception as e:
            print(f"An error occurred during the experiment: {e}")
            running = False
        finally:
            pygame.quit()
            print("Experiment closed.")

    board.stop_stream()
    board.release_session()

    # Join process for sending data.
    print("Join process for sending data")
    data_logging_queue.put(None)
    logging_process.join()



import json
from time import time

import h5py
import numpy as np

from nubrain.experiment.global_config import GlobalConfig
from nubrain.image.tools import load_image_as_bytes, resize_image

global_config = GlobalConfig()


def eeg_data_logging(subprocess_params: dict):
    """
    Log experimental data. For now, save to local hdf file. TODO: Send to remote
    backend. To be run in separate process (using multiprocessing).
    """
    # ----------------------------------------------------------------------------------
    # *** Get parameters

    demo_mode = subprocess_params["demo_mode"]
    image_directory = subprocess_params["image_directory"]

    # EEG parameters
    eeg_board_description = subprocess_params["eeg_board_description"]
    eeg_sampling_rate = subprocess_params["eeg_sampling_rate"]
    n_channels_total = subprocess_params["n_channels_total"]
    eeg_channels = subprocess_params["eeg_channels"]
    marker_channel = subprocess_params["marker_channel"]
    eeg_channel_mapping = subprocess_params["eeg_channel_mapping"]
    eeg_device_address = subprocess_params["eeg_device_address"]

    # Timing parameters
    initial_rest_duration = subprocess_params["initial_rest_duration"]
    image_duration = subprocess_params["image_duration"]
    isi_duration = subprocess_params["isi_duration"]
    inter_block_grey_duration = subprocess_params["inter_block_grey_duration"]

    # Experiment structure
    n_blocks = subprocess_params["n_blocks"]
    images_per_block = subprocess_params["images_per_block"]

    # Filter parameters
    # bandstop_low_cutoff_freq = subprocess_params["bandstop_low_cutoff_freq"]
    # bandstop_high_cutoff_freq = subprocess_params["bandstop_high_cutoff_freq"]
    # bandstop_filter_order = subprocess_params["bandstop_filter_order"]
    # bandpass_low_cutoff_freq = subprocess_params["bandpass_low_cutoff_freq"]
    # bandpass_high_cutoff_freq = subprocess_params["bandpass_high_cutoff_freq"]
    # bandpass_filter_order = subprocess_params["bandpass_filter_order"]

    # nubrain_endpoint = subprocess_params["nubrain_endpoint"]
    # nubrain_api_key = subprocess_params["nubrain_api_key"]

    path_out_data = subprocess_params["path_out_data"]

    data_logging_queue = subprocess_params["data_logging_queue"]

    # ----------------------------------------------------------------------------------
    # *** Create and initialize HDF5 file

    experiment_metadata = {
        "config_version": global_config.config_version,
        "demo_mode": demo_mode,
        "image_directory": image_directory,
        "rest_condition_color": global_config.rest_condition_color,
        "stim_start_marker": global_config.stim_start_marker,
        "stim_end_marker": global_config.stim_end_marker,
        "hdf5_dtype": global_config.hdf5_dtype,
        "max_img_storage_dimension": global_config.max_img_storage_dimension,
        "experiment_start_time": time(),
        # EEG parameters
        "eeg_board_description": eeg_board_description,
        "eeg_sampling_rate": eeg_sampling_rate,
        "n_channels_total": n_channels_total,
        "eeg_channels": eeg_channels,
        "marker_channel": marker_channel,
        "eeg_channel_mapping": eeg_channel_mapping,
        "eeg_device_address": eeg_device_address,
        # Timing parameters
        "initial_rest_duration": initial_rest_duration,
        "image_duration": image_duration,
        "isi_duration": isi_duration,
        "inter_block_grey_duration": inter_block_grey_duration,
        # Experiment structure
        "n_blocks": n_blocks,
        "images_per_block": images_per_block,
        # Filter parameters
        # "bandstop_low_cutoff_freq": bandstop_low_cutoff_freq,
        # "bandstop_high_cutoff_freq": bandstop_high_cutoff_freq,
        # "bandstop_filter_order": bandstop_filter_order,
        # "bandpass_low_cutoff_freq": bandpass_low_cutoff_freq,
        # "bandpass_high_cutoff_freq": bandpass_high_cutoff_freq,
        # "bandpass_filter_order": bandpass_filter_order,
    }

    print(f"Initializing HDF5 file at: {path_out_data}")
    with h5py.File(path_out_data, "w") as file:
        # ------------------------------------------------------------------------------
        # *** Initialize hdf5 dataset for metadata

        # Create group for metadata.
        metadata_group = file.create_group("metadata")

        # Iterate over the Python dictionary and save each item as an attribute of the
        # "metadata" group.
        for key, value in experiment_metadata.items():
            # HDF5 attributes have limitations on data types. Complex types like
            # dictionaries or tuples are not natively supported. We check if the value
            # is a type that needs to be converted to a string. JSON is a convenient
            # format for this serialization.
            if isinstance(value, (dict, list, tuple)):
                # Serialize the complex type into a JSON string.
                metadata_group.attrs[key] = json.dumps(value)
            else:
                metadata_group.attrs[key] = value

        # ------------------------------------------------------------------------------
        # *** Initialize hdf5 dataset for EEG data

        # Initialize dataset for board data (EEG and additional channels). To handle a
        # variable number of timesteps, create a resizable dataset. We specify an
        # initial shape but set the 'maxshape' to allow one of the dimensions to be
        # unlimited (by setting it to None). 'chunks=True' is recommended for resizable
        # datasets for better performance. It lets h5py decide the chunk size.
        file.create_dataset(
            "board_data",
            shape=(n_channels_total, 0),  # Start with 0 timesteps
            maxshape=(n_channels_total, None),  # fixed_channels, unlimited_timesteps
            dtype=global_config.hdf5_dtype,
            chunks=True,
        )

        # ------------------------------------------------------------------------------
        # *** Initialize hdf5 dataset for stimulus data

        # Define the compound datatype for stimulus data. This is like defining the
        # columns of a table. Use a special vlen dtype for the variable-sized image
        # data.
        stimulus_dtype = np.dtype(
            [
                ("stimulus_start_time", np.uint64),
                ("stimulus_end_time", np.uint64),
                ("stimulus_duration_s", np.float32),
                ("image_file_path", h5py.string_dtype(encoding="utf-8")),
                ("image_category", h5py.string_dtype(encoding="utf-8")),
                # ("image_description", h5py.string_dtype(encoding="utf-8")),
                (
                    "image_bytes",
                    h5py.vlen_dtype(np.uint8),
                ),  # For variable-length byte arrays
            ]
        )

        n_images = n_blocks * images_per_block

        file.create_dataset(
            "stimulus_data",
            (n_images,),
            dtype=stimulus_dtype,
        )

    # ----------------------------------------------------------------------------------
    # *** Experiment loop

    stimulus_counter = 0

    while True:
        new_data = data_logging_queue.get(block=True)

        if new_data is None:
            # Received None. End process.
            print("Ending preprocessing & data saving process.")
            break

        new_board_data = new_data["board_data"]
        new_stimulus_data = new_data["stimulus_data"]

        with h5py.File(path_out_data, "a") as file:
            # --------------------------------------------------------------------------
            # *** Write board data to hdf5 file

            if new_board_data is not None:
                hdf5_board_data = file["board_data"]

                # Get the current number of samples in the dataset.
                n_existing_samples = hdf5_board_data.shape[1]
                # Get the number of samples in the new batch.
                n_new_samples = new_board_data.shape[1]

                # Resize the dataset to accommodate the new data.
                n_total_samples = n_existing_samples + n_new_samples
                hdf5_board_data.resize(n_total_samples, axis=1)

                # Write the new data batch into the newly allocated space.
                hdf5_board_data[:, n_existing_samples:n_total_samples] = new_board_data

            # --------------------------------------------------------------------------
            # *** Write image data to hdf5 file

            # It is possible to receive board data without stimulus metadata (e.g. for
            # inter-stimulus interval).
            if new_stimulus_data is not None:
                hdf5_stimulus_data = file["stimulus_data"]

                image_file_path = new_stimulus_data["image_file_path"]
                image_bytes = load_image_as_bytes(image_path=image_file_path)
                image_bytes = resize_image(image_bytes=image_bytes)

                stimulus_start_time = new_stimulus_data["stimulus_start_time"]
                stimulus_end_time = new_stimulus_data["stimulus_end_time"]
                stimulus_duration_s = new_stimulus_data["stimulus_duration_s"]
                image_file_path = new_stimulus_data["image_file_path"]
                image_category = new_stimulus_data["image_category"]
                # image_description = new_stimulus_data["image_description"]

                data_to_write = np.empty((1,), dtype=stimulus_dtype)
                data_to_write[0]["stimulus_start_time"] = stimulus_start_time
                data_to_write[0]["stimulus_end_time"] = stimulus_end_time
                data_to_write[0]["stimulus_duration_s"] = stimulus_duration_s
                data_to_write[0]["image_file_path"] = image_file_path
                data_to_write[0]["image_category"] = image_category
                # data_to_write[0]["image_description"] = image_description
                # The image data is stored as a numpy array of bytes (uint8).
                data_to_write[0]["image_bytes"] = np.frombuffer(
                    image_bytes,
                    dtype=np.uint8,
                )

                # Write the structured array to the dataset.
                hdf5_stimulus_data[stimulus_counter] = data_to_write

                print(f"Stimulus counter: {stimulus_counter}")
                stimulus_counter += 1

    # End of data preprocessing process.

    # TODO: Send data to cloud storage.
